<?php
/**
 * The model file of iclock module of ZDOO.
 *
 * @copyright   Copyright 2009-2020 青岛易软天创网络科技有限公司(QingDao Nature Easy Soft Network Technology Co,LTD, www.cnezsoft.com)
 * @license     ZPL (http://zpl.pub/page/zplv12.html)
 * @author      Dongdong Jia <jiadongdong@easycorp.ltd> 
 * @package     iclock
 * @version     $Id$
 * @link        http://www.zdoo.com
 */
class iclockModel extends model
{
    /**
     * Inital information exchange.
     *
     * @param  string    $SN
     * @access public
     * @return array
     */
    public function init($SN)
    {
        $initConfig = $this->config->iclock->init; 

        $response  = "GET OPTION FROM: $SN";
        $response .= "\nATTLOGStamp={$initConfig->attLogStamp}";
        $response .= "\nOPERLOGStamp={$initConfig->operLogStamp}";
        $response .= "\nErrorDelay={$initConfig->errorDelay}";
        $response .= "\nDelay={$initConfig->delay}";
        $response .= "\nTransTimes={$initConfig->transTimes}";
        $response .= "\nTransInterval={$initConfig->transInterval}";
        $response .= "\nTransFlag={$initConfig->transFlag}";
        $response .= "\nTimeZone={$initConfig->timeZone}";
        $response .= "\nRealtime={$initConfig->realTime}";
        $response .= "\nEncrypt={$initConfig->encrypt}";
        $response .= "\nServerVer={$initConfig->serverVer}";
        $response .= "\nPushProtVer={$initConfig->pushProtVer}";
        $response .= "\nPushOptionsFlag={$initConfig->pushOptionsFlag}";

        return array('result' => 'success', 'message' => $response); 
    } 

    /**
     * Get all attends from machine in the preiod. 
     * 
     * @param  string $period 
     * @param  string $SN 
     * @access public
     * @return void
     */
    public function getAttends($period, $SN = '')
    {
        if(empty($period)) return;

        $startTime = date("Y-m-d H:i:s",strtotime("-1 day"));
        $endTime = helper::now();
        switch($period)
        {
            case 'day':
                $startTime = date("Y-m-d H:i:s",strtotime("-1 day"));
                break;
            case 'week':
                $startTime = date("Y-m-d H:i:s",strtotime("-1 week"));
                break;
            case 'month':
                $startTime = date("Y-m-d H:i:s",strtotime("-1 month"));
                break;
            case 'quarter':
                $startTime = date("Y-m-d H:i:s",strtotime("-3 month"));
                break;
            case 'halfyear':
                $startTime = date("Y-m-d H:i:s",strtotime("-6 month"));
                break;
        }

        $SNArr = array_filter(array_unique(explode(',', $SN)));
        foreach($SNArr as $SN)
        {
            $this->createCMD($SN, "QUERY ATTLOG StartTime=$startTime\tEndTime=$endTime");
            $this->log("QUERY ATTLOG StartTime=$startTime\tEndTime=$endTime");
        }
    }

    /**
     * Query command from oa_cmd. 
     * 
     * @param  string    $SN 
     * @access public
     * @return string 
     */
    public function getCMDs($SN)
    {
        /** 1. Query all command from oa_cmd. */ 
        $commands = $this->dao->select('DISTINCT command')->from(TABLE_CMD)
            ->where('sn')->eq($SN)
            ->andWhere('response')->eq('')
            ->orderBy('createdDate desc')
            ->limit($this->config->iclock->cmdLimit)
            ->fetchPairs();

        if(empty($commands)) return '';
        
        /** 2. Clear attend info if delete user. */
        foreach($commands as $command)
        {
            if(strpos($command, 'DELETE USERINFO PIN=') === false) continue;
            
            $pin = explode('=', $command)[1];
            $this->createCMD($SN, "DELETE ATTLOG PIN=$pin");
        }
    
        $requests = array();
        foreach($commands as $command) $requests[] = $command; 
        $requests = implode("\n", $requests);

        return $requests;
    }

    /**
     * Generate cmd to attend machine. 
     * 
     * @param  string    $SN 
     * @param  string    $command 
     * @access public
     * @return void
     */
    public function createCMD($SN, $command)
    {
        $cmd = new stdClass();
        $cmd->id          = uniqid() . mt_rand(1,100000); 
        $cmd->sn          = $SN;
        $cmd->command     = "C:$cmd->id:DATA " . $command;  
        $cmd->createdDate = helper::now();
        
        $this->dao->insert(TABLE_CMD)->data($cmd)->exec();
    }

    /**
     * Delete request command . 
     *  通用错误码  描述 
     *   0          成功 
     *  -1          参数错误 
     *  -3          读写错误 
     *  -10         设备中不存在PIN所指定的用户 
     *  -1001       容量限制
     * 
     * @param  string    $SN 
     * @param  string    $cmdData 
     * @access public
     * @return void
     */
    public function deleteCMD($SN, $cmdData)
    {
        $deleteIds = array();

        $cmdData = explode("\n", $cmdData);
        foreach($cmdData as $cmd) 
        {
            if(empty(trim($cmd))) continue;
            
            parse_str($cmd, $param);

            if($param['Return'] == '0')
            {
                $deleteIds[] = $param['ID'];
            }
            else
            {
                /* Update reponse when return code is not 0. */
                $this->dao->update(TABLE_CMD)
                    ->set('response')->eq($param['Return'])
                    ->where('id')->eq($param['ID'])
                    ->exec();
            }
        }

        /* Batch delete cmd request when return code is 0. */
        if(!empty($deleteIds))
        {
            $this->dao->delete()->from(TABLE_CMD)->where('id')->in($deleteIds)->exec();
            $this->log('deleted ids:' . implode(",", $deleteIds));
        }
    }

    /**
     * Sync zdoo users to attend machine. 
     * 
     * @param  string    $SN 
     * @access public
     * @return void
     */
    public function syncUsers2Machine($SN)
    {
        $users = $this->dao->select('account, realname')->from(TABLE_USER)
            ->where('deleted')->eq('0')
            ->andWhere('locked')->ne('2199-12-31 00:00:00')
            ->fetchPairs();

        foreach($users as $account => $realname) $this->addUser($account, $realname, $SN);
    }

    /**
     * Sync attend records of machine to zdoo. 
     * 
     * @param  string    $SN 
     * @access public
     * @return array 
     */
    public function syncAttend2zdoo($SN)
    {
        $attendLogs = file_get_contents("php://input");
        if(empty($attendLogs)) return array('result' => 'success', 'message' => 'OK');
        $this->log('', '', $attendLogs);

        $users = $this->dao->select('account, realname')->from(TABLE_USER)->fetchPairs();

        $successNum = 0;
        $userLogs   = array();
        $attendLogs = explode("\n", $attendLogs);
        foreach($attendLogs as $attendLog) 
        {
            if(empty(trim($attendLog))) continue;

            /** 1. Sync the attend data of machine to oa_attlog. */
            $attendLog = str_replace("\t", "_", $attendLog);
            $attendLog = explode('_', $attendLog);

            $log = new stdClass();
            $log->sn         = $SN;
            $log->userPin    = $attendLog[0];
            $log->verifyTime = $attendLog[1];
            $log->status     = $attendLog[2];
            $log->verifyType = $attendLog[3];
            $log->workCodeId = $attendLog[4];

            $this->dao->replace(TABLE_ATTLOG)->data($log)->exec();
            if(dao::isError()) continue;

            $successNum++;

            /** 2.Update oa_attend of zdoo. */
			if(!isset($users[$log->userPin])) continue;

            list($date, $time) = explode(' ', $log->verifyTime);

            if(isset($userLogs[$log->userPin][$date]))
            {
                if($time < $userLogs[$log->userPin][$date]['signIn']) $userLogs[$log->userPin][$date]['signIn']   = $time;
                if($time > $userLogs[$log->userPin][$date]['signOut']) $userLogs[$log->userPin][$date]['signOut'] = $time;
            }
            else
            {
                $userLogs[$log->userPin][$date]['signIn']  = $time;
                $userLogs[$log->userPin][$date]['signOut'] = $time;
            }
        }

        $result = $this->updateAttend($userLogs); 

        $response = "OK:$successNum";

        return array('result' => 'success', 'message' => $response); 
    } 

    /**
     * Sync operate log. 
     * 
     * @param  string    $SN 
     * @access public
     * @return array 
     */
    public function syncOperate2zdoo($SN)
    {
        $dataPack = file_get_contents("php://input");
        if(empty($dataPack) || strpos($dataPack, 'OPLOG') !== 0) return array('result' => 'success', 'message' => 'OK');

        $this->log('', '', $dataPack);

        $dataPack = explode("\n", $dataPack);
        $successNum = 0;
        foreach($dataPack as $data) 
        {
            if(empty(trim($data))) continue;
            if(strpos($data, 'OPLOG') !== 0) continue;
            
            $data = str_replace("OPLOG ", "", $data);
            $data = str_replace("\t", "_", $data);
            $data = explode('_', $data);

            $operate = new stdClass();
            $operate->sn       = $SN;
            $operate->opType   = $data[0];
            $operate->operator = $data[1];
            $operate->opTime   = $data[2];
            $operate->opWho    = $data[3];
            $operate->value1   = $data[4];
            $operate->value2   = $data[5];
            $operate->value3   = $data[6];

            $this->dao->replace(TABLE_OPLOG)->data($operate)->exec();
            if(!dao::isError()) $successNum++;
        } 

        $response = "OK:$successNum";
        return array('result' => 'success', 'message' => $response); 
    } 

    /**
     * Update oa_attend table.
     *
     * @param  array     $userLogs
     * @access public
     * @return bool
     */
    public function updateAttend($userLogs)
    {
        $this->loadModel('attend');

        $dates = array();
        foreach($userLogs as $account => $logs) $dates = array_merge($dates, array_keys($logs));

        $attends = $this->dao->select('*')->from(TABLE_ATTEND)
			->where('account')->in(array_keys($userLogs))
			->andWhere('`date`')->in($dates)
			->fetchGroup('account', 'date');

        foreach($userLogs as $account => $logs)
        {
            foreach($logs as $date => $log)
            {
                if(isset($attends[$account][$date]))
                {
                    $attend = $attends[$account][$date];

                    if($attend->signIn < $log['signIn'] && $attend->signOut > $log['signOut']) continue;

                    if($attend->signIn  > $log['signIn'])  $attend->signIn  = $log['signIn'];
                    if($attend->signOut < $log['signOut']) $attend->signOut = $log['signOut'];
                    $attend->status = $this->attend->computeStatus($attend);

                    $this->dao->update(TABLE_ATTEND)->data($attend)->where('id')->eq($attend->id)->exec();
                }
                else
                {
                    $attend = new stdclass();
                    $attend->account = $account;
                    $attend->date    = $date;
                    $attend->signIn  = $log['signIn'];
                    $attend->signOut = $log['signOut'];
                    $attend->ip      = helper::getRemoteIp();
                    $attend->device  = 'iclock';
                    $attend->client  = 'iclock';
                    $attend->status  = $this->attend->computeStatus($attend);

                    $this->dao->insert(TABLE_ATTEND)->data($attend)->autoCheck()->exec();
                }
            }
        }

        return !dao::isError();
    }

    /**
     * Add user to machine.
     * 
     * @param  string   $pin 
     * @param  string   $name 
     * @param  string   $SN 
     * @access public
     * @return void
     */
    public function addUser($pin, $name, $SN = '')
    {
        if(empty($SN)) $SN = $this->config->iclock->sn; 
        $SNArr = array_filter(array_unique(explode(',', $SN)));

        foreach($SNArr as $SN)
        {
            $this->createCMD($SN, "UPDATE USERINFO PIN=$pin\tName=$name");
            $this->log("Add or edit user to machine: $name");
        }
    }

    /**
     * Delete user from attend machine. 
     * 
     * @param  string    $pin 
     * @param  string    $SN 
     * @access public
     * @return void
     */
    public function deleteUser($pin, $SN = '')
    {
        if(empty($SN)) $SN = $this->config->iclock->sn; 
        $SNArr = array_filter(array_unique(explode(',', $SN)));

        $attend = $this->dao->select('date, signOut')->from(TABLE_ATTEND)->where('account')->eq($pin)->orderBy('`date` desc')->limit(1)->fetch();

        foreach($SNArr as $SN)
        {
            /* 1.Query attend for sync to zdoo. */
            $startTime = empty($attend) ? "2020-01-01 00:00:00" : $attend->date . ' ' . $attend->signOut;
            $endTime = helper::now();
            $this->createCMD($SN, "QUERY ATTLOG PIN=$pin\tStartTime=$startTime\tEndTime=$endTime");

            /* 2.Delete user in machine. */
            $this->createCMD($SN, "DELETE USERINFO PIN=$pin");
            $this->log("Delete user from machine: $pin");
        }
    }

    /**
     * Tak notes the iclock log to tmp/log/iclock.xxx.log 
     * 
     * @param  string    $data 
     * @param  string    $result 
     * @param  string    $errors 
     * @access public
     * @return void
     */
    public function log($data = '', $result = '', $errors= '')
    {
        commonModel::log($this->app->getURI(), $data, $result, $errors, $logFile = 'iclock');
    }
}
